"""
This module is used to create the api-documentation-creation in swagger.
The function :func:`ourobori.docs.swagger.create_swagger_definition`
is used in your ``app.py`` to create the documentation.
The required definition of your APISPECS in your ``service.py`` can be created
using the classes :class:`ourobori.docs.swagger.APISpec` and
:class:`ourobori.docs.swagger.OutputDefinition`.
If you defined everything correctly the api-documentation is available (after
you started your service) at
`http://<YOUR_HOST>:<YOUR_PORT>/api/doc`.
"""

from argparse import Namespace
from pathlib import Path
from typing import List
from typing import NoReturn
from typing import Optional
from typing import Union

import attr
from pydantic import BaseModel
import yaml


@attr.s(auto_attribs=True)
class OutputDefinition:
    """
    Class representing an additional output-definition for an
    api-documentation.
    The definition includes the http-status-code of the output
    (``status_code``), the description for this output (``description``)
    and the content-form of the output (``schema``).

    Attributes:
        status_code: The http-status-code of the output.
        description: The description for the ouput.
        schema: The content-form of the output.
    """
    status_code: str
    description: str
    schema: BaseModel


@attr.s(auto_attribs=True)
class Contact:
    """
    Class representing a contact to be used as parameter ``contact`` for
    :func:`create_swagger_definition`.

    Attributes:
        name: The full name of the contact-person
        email: The email of the contact-person
        url: The url for the service
    """
    name: str
    email: str
    url: str


@attr.s(auto_attribs=True)
class License:
    """
    Class representing a license to be used as parameter ``license`` for
    :func:`create_swagger_definition`.

    Attributes:
        name: The name of the license
        url: The url with license-description
    """
    name: str
    url: str


@attr.s(auto_attribs=True)
class APISpec:
    """
    Class representing the definition of an api to create a
    swagger-documentation for.

    Attributes:
        api_url: The route of the api.
        description: The description for the api.
        operation_id: The operation-id for the api. Should be in the form
            ``api.XXX`` but does nothing special.
        input_description: The description of the input required by the api.
        input_schema: The schema used for parsing of the request-parameters.
        output_description: The description of the output created by the api.
        output_schema: The schema used for parsing the output of the
            response of the api.
        additional_output_definitions: Additional output-definitions, for
            example the possible errors which could be occur calling the api.
    """
    api_url: str
    description: str
    input_description: str
    input_schema: BaseModel
    output_description: str
    output_schema: BaseModel
    additional_output_definitions: Optional[List[OutputDefinition]] = None

    def get_def(self) -> List[dict]:
        """
        Combines all relevant informations of the apispec to create a
        swagger-documentation in :func:`create_swagger_definition`.

        Returns:
            path, definitions: The combined informations required in the form
                :func:`create_swagger_definition` needs them to create a
                swagger-documentation.
        """
        definitions = dict()
        for schema in [self.input_schema, self.output_schema]:
            definition = schema.schema().get('definitions')
            if definition:
                definitions.update(definition)

        post = {'description': self.description,
                'operationId': f'api.{self.api_url.replace("/", ".")}',
                'produces': ['application/json'],
                'parameters': [{'in': 'body',
                                'name': 'body',
                                'description': self.input_description,
                                'required': True,
                                'schema': self.input_schema.schema()}],
                'responses': {'200': {'description': self.output_description,
                                      'schema': self.output_schema.schema()}}}

        if self.additional_output_definitions:
            assert isinstance(self.additional_output_definitions, list)
            for response in self.additional_output_definitions:
                response_schema = response.schema.schema()
                post['responses'].update({
                    response.status_code: {
                        'description': response.description,
                        'schema': response_schema}})

                additional_definitions = response_schema.get('definitions')
                if additional_definitions:
                    definitions.update(additional_definitions)

        path = {self.api_url: {'post': post}}
        return path, definitions


def create_swagger_definition(rules,
                              options: Namespace,
                              api_specs: List[APISpec]) -> str:
    """
    Creates a swagger-definition based on the input-parameters and stores it
    in the ``destination_file``.

    Note:
        This function creates a local file ``apispec.yml`` in the current
        working directory.

    Parameters:
        protocol: The protocol of the service. Should be either ``'http'``
            or ``'https'``.
        version: The version of the service.
        title: The title to use on the documentation-page.
        description: The description to show on the documentation-page.
        contact: The contact for the service.
        hostname: The hostname where the service is run.
        port: The port the service is listening to.
        api_specs: The apispecs to use to create the swagger-documentation.
        destination_file: The filepath where to store the
            swagger-definition-file.
        license: The license the service is licensed at.

    Returns:
        swagger_definition_path: The location of the swagger-definition-file.
            This path should be used to pass to
            :func:`ourobori.apps.http.run_service` inside the ``app.py`` of the
            service.
    """
    port = options.port or rules.port
    hostname = options.hostname
    protocol = 'http'

    destination_file = (Path.cwd() / 'apispec.yml').absolute()
    swagger_def = dict(swagger='2.0',
                       basePath='/',
                       info=dict(version=rules.version,
                                 title=f'{rules.servicename} API-documentation',
                                 description=rules.description,
                                 contact=rules.contact.__dict__,
                                 license=rules.license.__dict__),
                       servers=[dict(url=f'{protocol}://{hostname}:{port}/')],
                       paths=dict(),
                       definitions=dict())
    for apispec in api_specs:
        path, definitions = apispec.get_def()
        swagger_def['paths'].update(path)
        swagger_def['definitions'].update(definitions)

    destination_file.write_text(yaml.dump(swagger_def,
                                          default_flow_style=False,
                                          explicit_start=True))
    swagger_definition_path = str(destination_file.absolute())
    return swagger_definition_path


__all__ = ['APISpec',
           'Contact',
           'create_swagger_definition',
           'License',
           'OutputDefinition']
